/*
 *  permute. generate all permutations from lines of a text file
 *  Copyright (C) 2000    joe bentley
 *  rbentley@pigseye.kennesaw.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <assert.h>
#include <iostream.h>
#include <string.h>

//=============================================================================
//
//	globals
//
//=============================================================================

unsigned	* the_factorial_buf = 0,
		the_factorial_buf_size = 0;

const char * const	the_help_text[ ] =
{
	"permute -- generate all possible permutations of line from a text file",
	"syntax: permute [ -r | --reverse ] [ -s | --spaces ] < infile > outfile",
	"    or: permute [ -h | --help | -? | ? ]",
	"    or: permute [ -w | --warranty ]",
	0
};

const char * const	the_copyright_text[ ] =
{
	"permute version 1.0, copyright (C) 2000  joe bentley",
	0
};

const char * const	the_warranty_notice_text[ ] =
{
	"permute comes with ABSOLUTELY NO WARRANTY; for details type",
	"   `permute --warranty'.",
	0
};

const char * const	the_warranty_text[ ] =
{
	"This program is free software; you can redistribute it and/or modify",
	"it under the terms of the GNU General Public License as published by",
	"the Free Software Foundation; either version 2 of the License, or",
	"(at your option) any later version.",
	"",
	"This program is distributed in the hope that it will be useful,",
	"but WITHOUT ANY WARRANTY; without even the implied warranty of",
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
	"GNU General Public License for more details.",
	"",
	"You should have received a copy of the GNU General Public License",
	"along with this program; if not, write to the Free Software",
	"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA",
	0
};


//=============================================================================
//
//	generic useful templates
//
//=============================================================================

template < class data_type >
data_type min( data_type left, data_type right )
{
	return left < right ? left : right;
}

template < class data_type >
data_type max( data_type left, data_type right )
{
	return left > right ? left : right;
}

template < class data_type >
void dump_buf( const data_type * buf, unsigned buf_size )
{
	if( buf_size-- )
	{
		cout << *buf++;
		while( buf_size-- )
			cout << ", " << *buf++;
	}
}

template < class data_type >
unsigned count_array( const data_type * buf )
{
	unsigned	count = 0;
	while( *buf++ )
		count++;
	return count;
}

//=============================================================================
//
//	printable is an abstract class that provides printing behavior
//
//=============================================================================
class printable
{
public:
virtual	void print( ostream & ) const = 0;
};

inline ostream & operator << ( ostream & o, const printable & p )
{
	p.print( o );
	return o;
}

//=============================================================================
//
//	pset has 256 bits and standard set semantics
//
//=============================================================================
class pset : public printable
{
public:
	void print( ostream & ) const;

	void make_empty( );
	void make_full( );

	void set( unsigned );
	void clear( unsigned );
	int test( unsigned ) const;

	unsigned get_first( void ) const;
	unsigned get_next( unsigned ) const;
	unsigned get_nth( unsigned ) const;
	unsigned size( void ) const { return 32 * 8; }

protected:
	unsigned byte_number( unsigned element ) const;
	unsigned compute_mask( unsigned element ) const;

private:
	char	data[ 32 ];
};

void pset::print( ostream & o ) const
{
	o << "[ ";
	unsigned	element = get_first( );
	if( element < size() )
	{
		o << element << ' ';
		element = get_next( element );
		while( element < size() )
		{
			o << element << ' ';
			element = get_next( element );
		}
	}
	o << ']';
}

unsigned pset::byte_number( unsigned element ) const
{
	return element >> 3;
}

unsigned pset::compute_mask( unsigned element ) const
{
	return 1 << ( element & 7 );
}

void pset::make_empty( )
{
	memset( data, 0, sizeof( data ) );
}

void pset::make_full( )
{
	memset( data, -1, sizeof( data ) );
}

void pset::set( unsigned element )
{
	data[ byte_number( element ) ] |= compute_mask( element );
}

void pset::clear( unsigned element )
{
	data[ byte_number( element ) ] &= ~compute_mask( element );
}

int pset::test( unsigned element ) const
{
	return data[ byte_number( element ) ] & compute_mask( element );
}

unsigned pset::get_first( void ) const
{
	unsigned	element = 0;
	while( element < size() && !test( element ) )
		element++;
	return element;
}

unsigned pset::get_next( unsigned element ) const
{
	for( element++; element < size( ); element++ )
		if( test( element ) )
			break;
	return element;
}

unsigned pset::get_nth( unsigned n ) const
{
	unsigned	element = get_first( );
	while( n-- )
		element = get_next( element );
	return element;
}

//=============================================================================
//
//	textfile is a simple class to read in a text
//		file and convert to char*[]
//
//=============================================================================
class text_file
{
public:
	text_file( istream &, unsigned buf_size = 1024 );
virtual ~text_file( );

	char	** fix( void ) const;
	void reset( void );

protected:
	class text_file_line
	{
	public:
		text_file_line( const char * );
	virtual	~text_file_line( );
		unsigned height( void ) const;

		char	* data;
		text_file_line	* next;
	};

private:
	text_file_line	* root;
};

unsigned text_file::text_file_line::height( void ) const
{
	unsigned	count = 1;
	const text_file_line	* l = next;
	while( l )
	{
		l = l->next;
		count++;
	}
	return count;
}

text_file::text_file_line::text_file_line( const char * s )
{
	unsigned	data_size = strlen( s ) + 1;
	data = new char [ data_size + 1 ];
	strcpy( data, s );
	next = 0;
}

text_file::text_file_line::~text_file_line( )
{
	if( next )
		delete next;
}

text_file::~text_file( )
{
	reset( );
}

void text_file::reset( void )
{
	if( root )
		delete root;
	root = 0;
}

char ** text_file::fix( void ) const
{
	unsigned	table_size = root ? root->height( ) + 1 : 1;
	char	** table = new char * [ table_size + 1 ],
		** ptr = table;
	assert( table );
	text_file_line	* l = root;
	while( l )
	{
		*ptr++ = l->data;
		l = l->next;
	}
	*ptr++ = 0;
	return table;
}

text_file::text_file( istream & infile, unsigned buf_size )
{
	root = 0;
	char	* buf = new char [ buf_size ];
	assert( buf );
	infile.getline( buf, buf_size );
	if( infile.good( ) )
	{
		root = new text_file_line( buf );
		text_file_line	* tail = root;
		assert( root );
		infile.getline( buf, buf_size );
		while( infile.good( ) )
		{
			text_file_line	* l = new text_file_line( buf );
			assert( l );
			tail->next = l;
			tail = l;
			infile.getline( buf, buf_size );
		}
	}
	delete[] buf;
}

char ** read_text_file( istream & infile )
{
	text_file	f( infile );
	char	** data = f.fix( );
	return data;
}

//=============================================================================
//
//	some factorial routines. note that they require the globals
//		the_factorial_buf and the_factorial_buf_size defined above
//
//=============================================================================

unsigned factorial( unsigned argument )
{
	assert( argument < the_factorial_buf_size );
	return the_factorial_buf[ argument ];
}

unsigned get_max_factorial_argument( void )
{
	unsigned	current = 1,
			previous = 0,
			count = 1;
	while( current >= previous )		// intentional overflow. yikes
	{
		count++;
		previous = current;
		current = current * ( count + 1 );
	}
	return count;
}

void kill_factorial( void )
{
	if( the_factorial_buf )
		delete[] the_factorial_buf;
}

void init_factorial( void )
{
	unsigned	buf_size = get_max_factorial_argument( ),
			running_product = 1;
	the_factorial_buf = new unsigned [ buf_size ];
	assert( the_factorial_buf );
	the_factorial_buf_size = buf_size;
	for( unsigned counter = 0; counter < buf_size; counter++ )
	{
		the_factorial_buf[ counter ] = running_product;
		running_product *= counter + 1;
	}
}

//=============================================================================
//
//	this is the primary routine.  everything else
//		just reads stdin or sets up options
//
//=============================================================================

void put_permutation( unsigned * buf, unsigned buf_size, unsigned perm_index )
{
	pset	p;
	p.make_full( );
	while( buf_size > 1 )
	{
		unsigned	block_size = factorial( buf_size - 1 ),
				i = p.get_nth( perm_index / block_size );
		*buf++ = i;
		p.clear( i );
		perm_index %= block_size;
		buf_size--;
	}
	*buf++ = p.get_first( );
}

//=============================================================================
//
//	options and set up code
//
//=============================================================================

void dump_string_list( ostream & outfile, const char * const * sl )
{
	while( *sl )
		outfile << *sl++ << endl;
}

void show_warranty( void )
{
	dump_string_list( cout, the_copyright_text );
	cout << endl;
	dump_string_list( cout, the_warranty_text );
}

void do_help( void )
{
	dump_string_list( cout, the_copyright_text );
	cout << endl;
	dump_string_list( cout, the_help_text );
	cout << endl;
	dump_string_list( cout, the_warranty_notice_text );
}

void do_error_option( const char * s )
{
	cerr << "unknown option " << s << endl;
}

int is_help_option( const char * s )
{
	return	!strcasecmp( s, "--help" ) || !strcasecmp( s, "-h" )
		|| !strcmp( s, "-?" ) || !strcmp( s, "?" );
}

int is_reverse_option( const char * s )
{
	return !strcmp( s, "--reverse" ) || !strcmp( s, "-r" );
}

int is_spaces_option( const char * s )
{
	return !strcmp( s, "--spaces" ) || !strcmp( s, "-s" );
}

int is_warranty_option( const char * s )
{
	return !strcasecmp( s, "--warranty" ) || !strcasecmp( s, "-w" );
}

int get_options(
	const char * const * arg_strings, unsigned count,
	int * step_value, int * separate_with_spaces_flag )
{
	*step_value = 1;
	*separate_with_spaces_flag = 0;

	int	show_help_flag = 0,
		show_warranty_flag = 0,
		err = 0;
	for( unsigned string_index = 1; string_index < count; string_index++ )
	{
		const char	* s = arg_strings[ string_index ];
		if( is_reverse_option( s ) )
			*step_value = -1;
		else if( is_spaces_option( s ) )
			*separate_with_spaces_flag = 1;
		else if( is_help_option( s ) )
			show_help_flag = 1;
		else if( is_warranty_option( s ) )
			show_warranty_flag = 1;
		else
		{
			do_error_option( s );
			err = 1;
		}
	}
	if( show_warranty_flag )
		show_warranty( );
	else if( show_help_flag )
		do_help( );
	else if( err )
		cerr << "Try `permute --help' for more information" << endl;
	return show_warranty_flag || show_help_flag || err;
}

void output_element( const char * s, int leading_space )
{
	if( leading_space )
		cout << ' ';
	cout << s;
}

void do_output(
	const char * const * string_list,
	unsigned * string_index_buf, unsigned element_count,
	int use_spaces_flag )
{
	unsigned	counter = 0;
	const char	* s = string_list[ string_index_buf[ counter++ ] ];
	output_element( s, 0 );
	while( counter < element_count )
	{
		s = string_list[ string_index_buf[ counter++ ] ];
		output_element( s, use_spaces_flag );
	}
	cout << endl;
}

void setup_perm_bounds(
	int up_or_down_flag,
	unsigned element_count,
	unsigned * lower, unsigned * upper )
{
	if( up_or_down_flag > 0 )
	{
		*lower = 0;
		*upper = factorial( element_count ) - 1;
	}
	else
	{
		*lower = factorial( element_count ) - 1;
		*upper = 0;
	}
}

void do_permute(
	const char * const * buf, unsigned buf_size,
	int counter_step, int use_spaces_flag )
{
	unsigned	counter,
			target,
			* perm_buf = new unsigned [ buf_size + 1 ];
	assert( perm_buf );
	setup_perm_bounds( counter_step, buf_size, &counter, &target );
	put_permutation( perm_buf, buf_size, counter );
	do_output( buf, perm_buf, buf_size, use_spaces_flag );
	while( counter != target )
	{
		counter += counter_step;
		put_permutation( perm_buf, buf_size, counter );
		do_output( buf, perm_buf, buf_size, use_spaces_flag );
	}
	delete[] perm_buf;
}

int main( int argc, const char * const * argv )
{
	int	delta = 1,
		use_spaces_flag = 0;
	if( !get_options( argv, argc, &delta, &use_spaces_flag ) )
	{
		init_factorial( );
		char	** buf = read_text_file( cin );
		unsigned	buf_size = count_array( buf ),
				max_buf_size;
		max_buf_size = get_max_factorial_argument( );
		if( buf_size && buf_size < max_buf_size )
			do_permute( buf, buf_size, delta, use_spaces_flag );
		else if( buf_size )
			cerr	<< "Unable to permute more than "
				<< max_buf_size - 1 << " elements" << endl
				<< "and even that would take a _really_ long"
				<< " time" << endl
				<< "If you're looking for a way to solve"
				<< " word jumbles I recommend `wordplay'"
				<< endl;
		kill_factorial( );
	}
	return 0;
}
